import re, json 
from dataclasses import dataclass, asdict as dataclasses_asdict 
from dataclasses_json import dataclass_json

from typing import Callable

from zohavi.zbase.staple import ZStaple
from mclogger import MCLogger

class WebField( ZStaple ):
	def __init__(self, tables, app=None, logger=None):
		super().__init__(app = app, logger = logger )
		self.debug = True

		self._tables = tables
		self._table_lookup = self._calc_web_table_lookup()

		self.log_debug(f"Stored: {json.dumps(tables) }")

	def get_field(self, field_name):
		# for field in self._fields:
		# 	if field['id'] == field_name: return field
		return self._fields[field_name]
 
	def _calc_web_table_lookup(self):
		table_lookup = {}
		for index, table_rec in enumerate( self._tables ):
		#create new dict entry for each of the web fields
			for web_field_name in table_rec['fields'].keys():
				table_lookup[ web_field_name ] =  table_rec 
		return table_lookup

	#############################################################################################################
	#Validate form data in json format
	@MCLogger.logfunc_cls('logger')
	def validate(self, data):
		# validation_ok = True 
		result = { 'success': True, 'validations':[] }	#Assume successful

		self.log_debug( "validating:" + json.dumps(data ) )
		self.log_debug( "Master:" + json.dumps(self._tables ) )

		for data_row in data:	#loop through each web data field
			if not 'id' in data_row: raise Exception( f"Missing field [id] in {data_row}.  Json format shoudl have 'id' and 'value' elements")

			web_field_name = data_row['id']
			# breakpoint()
			if not web_field_name in self._table_lookup: raise Exception( f"[{web_field_name}] from submit data not found in schema { self._tables }")
			# table_ref = self._field_lookup[ web_field_name ]
			
			#Use the reverse lookup to find the table associated with the field
			field_info = self._table_lookup[ web_field_name ]['fields'][ web_field_name ]

			if field_info.get('validation', False):  #has validation
				self.log_debug( f" Checking validation {web_field_name}=>[{data_row['value']}] Rule:{json.dumps( field_info['validation']) } ") 
				# breakpoint()
				result = self._validate_run_validation_rule(web_field_name, data_row['value'] , field_info['validation'], result  )

			if field_info.get('transform', False):  #has transformation
				data_row['value'] = self._run_transformation( data_row['value'] , field_info['transform'] )
		
		return result


	#############################################################################################################
	#
	def _run_transformation(self, data_value, transform_rules):
		self.log_debug( f" data_value={data_value}; transform={json.dumps(transform_rules) }")

		ret_value = data_value
		for rule in transform_rules: 
			func = getattr(Transform, rule, Transform._func_not_found)  
			ret_value = func( ret_value , transform_rules[rule] )

		return ret_value


	#############################################################################################################
	#Run the actual validation rule
	def _validate_run_validation_rule(self, web_field_name, data_value, validation_rules, validation_results:dict ):
		result = validation_results
		if not result: result = { 'success': True, 'validations':[] }
		# validation_ok = True 
		# if not data_value: breakpoint()
		# breakpoint()
		self.log_debug( f" data_value={data_value}; validation_rule={json.dumps(validation_rules)}")
		if not data_value and 'required' in validation_rules.keys() and validation_rules['required'] == False: return result
		for rule in validation_rules: 
			func = getattr(Validate, rule, Validate._func_not_found)  
			ret_result = func( data_value , validation_rules[rule], self.log_error )	#Pass reference to error message for failed validatinos
			ret_result.web_field_name = web_field_name
			result['validations'].append(  dataclasses_asdict(ret_result ) )
			log_message = f"Validation check [{rule}:{validation_rules[rule]}] on [{data_value}] => {ret_result}"
			# if ret_check:
			# 	self.log_debug( log_message ) 
			# else:
			# 	validate_fail_reason['01'] = log_message
			# 	self.log_error( log_message )

			result['success'] = result['success'] and ret_result.success

		self.log_debug("Returning validation check from :" + json.dumps(validation_rules) + " => " + str(result['success']) )
		return result
 


class Transform(ZStaple):

	@staticmethod
	def _func_not_found( value, param) :
		Transform.log_error("transformation function not found")

	#############################################################################################################
	#Check if the current 
	@staticmethod
	def to_bool(value, param) :
		return Transform._map_value( value, {'':False, 'false':False, '0':False, 
											'True':True, 'true':True, '1':True, True:True} )
	@staticmethod
	def _map_value(value, map):
		return map.get( value, None)

@dataclass_json
@dataclass
class ValidateResult():
	web_field_name: str = "" 
	success: bool = False
	rule: str = ""
	value: object = None
	param: object = None
	err_no: str = ""
	err_msg: str = ""
	
	

class Validate(ZStaple):
	#############################################################################################################
	#Check if the current 
	@staticmethod
	def _get_data_field_field(  value, field_name):
		for data_fld in value:	#find the relevant data field in there
			if data_fld['id'] == field_name:
				return data_fld
		return None

	#############################################################################################################
	#validation function not found
	@staticmethod
	def _func_not_found( value, param, log_err_func: Callable =print ):
		vr = ValidateResult( rule='??', value=value,  param=param , err_no="VAL_010", err_msg='Validation Rule passed not found')
		Validate.log( vr.err_msg , log_err_func)
		return vr

	#############################################################################################################
	#Check that field is required 
	@staticmethod
	def required(  value, param, log_err_func: Callable =print ):  
		vr = ValidateResult( rule='required', value=value,  param=param )
		if not param: vr.success = True
		elif value: vr.success = True  
		else:  #e.g. parameter is present, and value not given
			vr.success = False  
			vr.err_no = "VAL_020"
			vr.err_msg = 'Required value not given'
			Validate.log( vr.err_msg , log_err_func)

		return vr 

	#############################################################################################################
	#Check that field has min length as required
	@staticmethod
	def text_min_len(  value, param, log_err_func: Callable =print  ): 
		vr = ValidateResult( rule='text_min_len', value=value,  param=param )

		if len( str( value  ) ) >= param: vr.success = True
		else: 
			vr.success = False 
			vr.err_no = "VAL_030"
			vr.err_msg = f"Value [{value}] should be minimum length of {param}, but actual length is {len( str( value  ) )}"
			Validate.log( vr.err_msg , log_err_func)
		return vr 

	#############################################################################################################
	#Check that field has max length as required
	@staticmethod
	def text_max_len(  value, param,log_err_func: Callable =print  ): 
		vr = ValidateResult( rule='text_max_len', value=value,  param=param )

		if len( str( value ) ) <= param: vr.success= True
		else: 
			vr.success = False 
			vr.err_no = "VAL_040"
			vr.err_msg = f"Value [{value}] should be max length of {param}, but actual length is {len( str( value  ) )}"
			Validate.log( vr.err_msg , log_err_func)
		
		return vr 

	#############################################################################################################
	#Check that number is gte
	@staticmethod
	def num_gte(   value, param, log_err_func: Callable =print ): 
		vr = ValidateResult( rule='num_gte', value=value,  param=param )
		if int( value ) >= int(param): vr.success= True
		else:
			vr.success = False 
			vr.err_no = "VAL_050"
			vr.err_msg = f"Value [{value}] should be greater than or equal to {param}, but actual is { int( value )}"
			Validate.log( vr.err_msg , log_err_func)
		
		return vr 

	#############################################################################################################
	#Check that number is gte
	def num_lte(  value, param , log_err_func: Callable =print  ): 
		vr = ValidateResult( rule='num_lte', value=value,  param=param )
		if int( value ) <= int(param): vr.success= True
		else:
			vr.success = False 
			vr.err_no = "VAL_060"
			vr.err_msg = f"Value [{value}] should be less than or equal to {param}, but actual is { int( value )}"
			Validate.log( vr.err_msg , log_err_func)
		return vr 

	#############################################################################################################
	#Check that field has max length as required
	@staticmethod
	def is_url(  value, param=None, log_err_func: Callable =print ):  
		vr = ValidateResult( rule='is_url', value=value,  param=param )
		url_re = r'(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:\'".,<>?\xab\xbb\u201c\u201d\u2018\u2019]))'
		url = re.findall(url_re, value )
		if url:  vr.success= True
		else:
			vr.success = False 
			vr.err_no = "VAL_070"
			vr.err_msg = f"[{value}] is not an URL"
			Validate.log( vr.err_msg , log_err_func)
		return vr 

	#############################################################################################################
	#Check that field has max length as required
	@staticmethod
	def is_ip(  value, param=None, log_err_func: Callable =print   ):  
		vr = ValidateResult( rule='is_ip', value=value,  param=param )
		url_re = r'^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$'
		url = re.findall(url_re, value )
		
		if url:  vr.success= True
		else:
			vr.success = False 
			vr.err_no = "VAL_080"
			vr.err_msg = f"[{value}] is not an IP"
			Validate.log( vr.err_msg , log_err_func)
		return vr 


	#############################################################################################################
	#Check that number is gte
	@staticmethod
	def is_unix_path(  value , param=None, log_err_func: Callable =print  ): 
		vr = ValidateResult( rule='is_unix_path', value=value,  param=param )
		path_re = r'\/?(\/?.+?)+[\/]?'
		path = re.findall(path_re, value )
		if path:  vr.success= True
		else:
			vr.success = False 
			vr.err_no = "VAL_090"
			vr.err_msg = f"[{value}] is not a unix-style path"
			Validate.log( vr.err_msg , log_err_func)
		return vr 


	#############################################################################################################
	#validation function not found
	@staticmethod
	def is_bool( value, param, log_err_func: Callable =print ):
		vr = ValidateResult( rule='required', value=value,  param=param )
		valid_values = [ True, False, '', None ]
		if value in valid_values:  vr.success= True
		else:
			vr.success = False 
			vr.err_no = "VAL_090"
			vr.err_msg = f"Value [{value}] is not a boolean"
			Validate.log( vr.err_msg , log_err_func)
		return vr 

	@staticmethod
	def log( message, log_err_func: Callable =print ):

		log_err_func( "Validation Failure: " + message )



"""empty message

Revision ID: 6250a38af3b8
Revises: 276c78c4a2b7
Create Date: 2022-04-18 02:03:38.122063

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '6250a38af3b8'
down_revision = '276c78c4a2b7'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('config_item')
    op.drop_table('config')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('config',
    sa.Column('sys_created_on', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('sys_changed_on', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('id', sa.INTEGER(), server_default=sa.text("nextval('config_id_seq'::regclass)"), autoincrement=True, nullable=False),
    sa.Column('area', sa.VARCHAR(length=25), autoincrement=False, nullable=True),
    sa.Column('category', sa.VARCHAR(length=25), autoincrement=False, nullable=True),
    sa.Column('sub_cat', sa.VARCHAR(length=25), autoincrement=False, nullable=True),
    sa.PrimaryKeyConstraint('id', name='config_pkey'),
    postgresql_ignore_search_path=False
    )
    op.create_table('config_item',
    sa.Column('sys_created_on', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('sys_changed_on', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('id', sa.INTEGER(), autoincrement=True, nullable=False),
    sa.Column('config_id', sa.INTEGER(), autoincrement=False, nullable=True),
    sa.Column('name', sa.VARCHAR(length=100), autoincrement=False, nullable=True),
    sa.Column('value', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
    sa.Column('env', sa.VARCHAR(length=10), autoincrement=False, nullable=True),
    sa.ForeignKeyConstraint(['config_id'], ['config.id'], name='config_item_config_id_fkey'),
    sa.PrimaryKeyConstraint('id', name='config_item_pkey')
    )
    # ### end Alembic commands ###



import json, datetime 

from sqlalchemy import DateTime, Column

from zohavi.zdb.models import TBL_Default
from zohavi.zbase.staple import ZStaple
from prepost import PreCond


############################################################################################################
############################################################################################################
class MDL_Config(ZStaple):
	def __init__(self, logger, db):
		super().__init__(logger=logger, app=None)
		self.db = db

		class TBL_Config(db.Model, TBL_Default):
			__tablename__ = 'config' 
			id = db.Column(db.Integer(), primary_key=True)
			area 		= db.Column(db.String(25) ) 
			category 	= db.Column(db.String(25) ) 
			sub_cat		= db.Column(db.String(25) ) 
			_config_items = db.relationship('TBL_ConfigItem', cascade='all,delete',  passive_deletes=True, back_populates="_config")
		
		class TBL_ConfigItem(db.Model, TBL_Default):
			__tablename__ = 'config_item' 
			id = db.Column(db.Integer(), primary_key=True)
			config_id = db.Column(db.Integer(), db.ForeignKey( 'config.id' , ondelete='CASCADE') ) 
			env =  db.Column(db.String(10) ) 
			name = db.Column(db.String(100) ) 
			value = db.Column(db.String(255) ) 		
			flask_name = db.Column(db.String(100) ) 
			_config = db.relationship('TBL_Config',   back_populates="_config_items")
	
		self._tables = {}
		self._tables['TBL_Config'] 	   = TBL_Config
		self._tables['TBL_ConfigItem'] = TBL_ConfigItem
	
	################################################################################################################
	def get_tables(self):
		return self._tables

	################################################################################################################
	@PreCond.matches_pattern( config_str='.*\/.*\/.*\/.*')
	def get_config(self, config_str):
		xpath = config_str.split('/')
		cls_tbl_item = self._tables['TBL_ConfigItem']
		cls_tbl_main = self._tables['TBL_Config']

		item_list = self.db.session.query( cls_tbl_item ).filter( cls_tbl_item.name == xpath[3] ).all()
		for item in item_list:
			if item._config.area == xpath[0] and item._config.category == xpath[1] and item._config.sub_cat == xpath[2]:
				return item
		return None

	################################################################################################################
	@PreCond.matches_pattern( config_str='.*\/.*\/.*\/.*')
	def set_config(self, config_str, value, flask_name = None):
		
		item = self.get_config( config_str = config_str )
		# 
		if item: item.value = value
		else:
			#check if the main entry exists
			xpath = config_str.split('/')
			tbl_main = self._set_config_main( xpath[0], xpath[1], xpath[2] )
			tbl_item = self._tables['TBL_ConfigItem']()
			tbl_item.name = tbl_item.flask_name = xpath[3]
			if flask_name: tbl_item.flask_name = flask_name
			tbl_item.value = value
			tbl_item._config = tbl_main
			self.db.session.add( tbl_item )
		self.db.session.commit()		 
	
	def _set_config_main(self, area, category, sub_cat):
		cls_tbl_main = self._tables['TBL_Config']
		config_main = self.db.session.query( cls_tbl_main ).filter( cls_tbl_main.area == area, cls_tbl_main.category == category, cls_tbl_main.sub_cat == sub_cat).first()
		if not config_main:
			config_main = cls_tbl_main()
			config_main.area = area
			config_main.category = category
			config_main.sub_cat = sub_cat
			self.db.session.add(config_main)
		return config_main
			


	def to_flask_config(self):
		config = {}

		cls_config_item = self._tables['TBL_ConfigItem']
		data_flask_config = self.db.session.query( cls_config_item ).filter( cls_config_item.flask_name != '' ).all()
		# breakpoint()
		for data_item in data_flask_config:
			config[ data_item.flask_name ]= data_item.value
		return config


	


import unittest
unittest.TestLoader.sortTestMethodsUsing = None

from pathlib import Path 
import sys, os, jsoncfg, json
sys.path.insert(0, '../../../')
 
# from zohavi.zcore.appcore import AppCore 

from flask import Flask, jsonify, render_template,send_file, request
from mclogger.mclogger import MCLogger

logger = MCLogger( 'test_log.txt' )
 

html = """<html>
			<body>
				<h1>hello world</h1>
			</body>
		</html>"""
site_update_schema = {
							"SiteEnv":[{ 	
								"module_name":"test_wc",
								"table_obj": "SiteEnv",
								"fields":{
											"si_env_id":{"field_db":"id",   "key":True},
											"si_env_name":{"field_db":"env_name", "validation":{"required":True, "text_min_len":3, "text_max_len":20} },
											"si_env_code":{"field_db":"env_code", "validation":{"required":True,  "text_max_len":10} },
											"si_env_desc":{"field_db":"env_desc", "validation":{"required":False,  "text_max_len":200} }
								}
							}],
							"SiteEnvSetting":[{ 	
								"module_name":"test_wc",
								"table_obj": "SiteEnv",
								"fields":{
											"si_env_id":{"field_db":"id",   "key":True},
											"si_env_cs":{"field_db":"env_cs", "validation":{"required":True} },
											"si_env_authlogon":{"field_db":"env_authlogon", "validation":{"required":True} },
											"si_env_authsite_name":{"field_db":"env_authsite_name", "validation":{"required":False,"text_min_len":4,   "text_max_len":10} },
											"si_env_authmethod":{"field_db":"env_auth_method", "validation":{"required":False} },
											
								}
							}]
					}

from sqlalchemy import Column, Integer, String 
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from flask_sqlalchemy import SQLAlchemy

from zohavi.zdb.models import TBL_Default
from zohavi.zwebui.data_ui_model import DataUIModel
from zohavi.zwebui.jinja_cust_funcs import JCFunc

Base = declarative_base()


	# cfg_main_runme 	= Column(String(10) ) #dev, qa, prod  
# 	_env_list = relationship("SiteEnv", backref=backref("site_env" ), lazy='select') 
	

# engine = create_engine('sqlite:///:memory:')
# Session = sessionmaker(bind=engine)
# session = Session()

# Base.metadata.create_all(engine)

# session.commit()

app = Flask(__name__)
app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///:memory:'
app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True

JCFunc.load_template_funcions( app.jinja_env)

# this variable, db, will be used for all SQLAlchemy commands
db = SQLAlchemy(app)

############################################################################################################
############################################################################################################
class SiteEnv( db.Model ,  TBL_Default):
	__tablename__ = 'site_env' 
	id = Column(Integer(), primary_key=True)
	env_name = Column(String(100) ) 
	env_desc = Column(String(200) ) 
	env_code = Column(String(5) )  
	env_cs = Column(String(5) )  
	env_authlogon = Column(String(5) )  
	env_authsite_name = Column(String(5) )  
	env_auth_method = Column(String(5) )  



db.create_all()
db.session.add( SiteEnv(env_name='dev', env_desc='Twitter Bot', env_code='dev1'  ) )
db.session.commit()

# myapp = AppCore(app, sccfg, 'dev' )

#################################################################
@app.route('/webui/<path:url>', methods=['GET' ])
def webui_path( url ):
	# path = current_app.config['ENV_BASE_DIR']
	search_file = os.getcwd()[:os.getcwd().rfind('/')] +"/" +url
	# breakpoint()

	if Path( search_file ).is_file():
		return send_file( search_file )

	print("file missing:" + str(search_file)   )
	abort(404)
	logger.debug(url)
	return render_template( 'test_table.html')

#################################################################
@app.route('/test/menu', methods=['GET' ] )
def test_menu( ):
	return render_template( 'test_menu.html')
 
#################################################################
@app.route('/test/table', methods=['GET' ] )
def test_table( ):
	return render_template( 'test_table.html')

#################################################################
@app.route('/test/button', methods=['GET' ] )
def test_button( ):
	return render_template( 'test_button.html')


#################################################################
@app.route('/button/click', methods=['POST' ] )
def test_button_click( ):
	return  json.dumps({'success':True}), 200


#################################################################
@app.route('/', methods=['GET' ] )
def test_all( ):
	routes = {}
	for rule_item in app.url_map._rules:
		routes[rule_item.rule] = {}
		routes[rule_item.rule]["functionName"] = rule_item.endpoint
		routes[rule_item.rule]["methods"] = list(rule_item.methods)

	return render_template( 'test_all.html', routes=routes)


#################################################################
@app.route('/test/group', methods=['GET' ] )
def test_group( ):
	data_ui = DataUIModel(  site_update_schema['SiteEnvSetting'] , db.session, logger )
	env_data = SiteEnv.query.first()
	return render_template( 'test_group.html', data_ui=data_ui, env_data=env_data)

#################################################################
@app.route('/test/group/save', methods=['POST' ] )
def test_group_save( ):
	data_ui = DataUIModel(  site_update_schema['SiteEnvSetting'] , db.session, logger )
	return data_ui.data_update_ajax( request.json )

#################################################################
@app.route("/test/table/ajax_add", methods=["POST"])
def test_table_add():
	logger.debug( request.json )
	data_ui = DataUIModel(  site_update_schema['SiteEnv'] , db.session, logger )
	return data_ui.data_update_ajax( request.json )
	# return json.dumps({'success':True}), 200

#################################################################
@app.route("/test/table/ajax_edit", methods=["POST"])
def test_table_edit():
	logger.debug( request.json )
	data_ui = DataUIModel( site_update_schema['SiteEnv'] , db.session, logger )
	return data_ui.data_update_ajax( request.json )

#################################################################
@app.route("/test/table/ajax_del", methods=["POST"])
def test_table_del():
	logger.debug( request.json )
	data_ui = DataUIModel(   site_update_schema['SiteEnv'] , db.session, logger )
	return data_ui.data_delete_ajax( request.json ) 



#################################################################
@app.route("/routes", methods=["GET"])
def getRoutes( ):
	routes = {}
	for rule_item in app.url_map._rules:
		routes[rule_item.rule] = {}
		routes[rule_item.rule]["functionName"] = rule_item.endpoint
		routes[rule_item.rule]["methods"] = list(rule_item.methods)
	return jsonify(routes)

 
 #################################################################
 #################################################################
if __name__ == '__main__':
	app.run(host="0.0.0.0", port=4601, debug=True)



import unittest
unittest.TestLoader.sortTestMethodsUsing = None

import sys, os, jsoncfg
sys.path.insert(0, '../../../')


from mclogger import MCLogger
from zohavi.zconfig.config_manager import CoreConfig 
from zohavi.zconfig.models import MDL_Config
from jsondbupload import JsonDBUpload

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_migrate import Migrate
 

sys_config1 = """{
					"dev":{
						"host":"0.0.0.0",
						"port":"4100" ,
						"log_file":"log_test_dev.txt",
						"db_core":"postgresql:///test1"
					},
					"qa":{
						"host":"0.0.0.0",
						"port":"4200" ,
						"log_file":"log_test_qa.txt",
						"db_core":"postgresql:///test1"
					},
					"prd":{
						"host":"0.0.0.0",
						"port":"4300" ,
						"log_file":"log_test_prd.txt",
						"db_core":"postgresql:///test1"
					},
			}"""

logger = MCLogger('log.txt')

########
sccfg = CoreConfig(config_str=sys_config1, base_dir=os.getcwd() )

# breakpoint()
app = Flask(__name__)
app.config['SQLALCHEMY_DATABASE_URI'] = sccfg.get_env_config('dev').db_core
db = SQLAlchemy(app)
migrate = Migrate(app, db)
cdb = MDL_Config( logger=None, db=db)
db.drop_all()
db.create_all()

class TestProc(unittest.TestCase):
	def test_0000_setup_app_config(self):
		pass
	
	def test_0010_setup_database(self):	
		JsonDBUpload(db, logger).update_tables_from_file( 'test_db_config_defaults.json') 

	
	def test_0020_load_core_config_str(self):
		sccfg = CoreConfig( config_str=sys_config1, base_dir=os.getcwd() )

		raw_cfg = jsoncfg.loads_config( sys_config1 )
		# breakpoint()
		logger.debug( os.getcwd() )
		self.assertEqual( sccfg.get_env_config('dev').env , 'dev'  )
		self.assertEqual( sccfg.get_env_config('dev').host , raw_cfg.dev.host()  )
		self.assertEqual( sccfg.get_env_config('dev').port , raw_cfg.dev.port()  )
		self.assertEqual( sccfg.get_env_config('dev').log_file , raw_cfg.dev.log_file()  )
		self.assertEqual( sccfg.get_env_config('dev').db_core , raw_cfg.dev.db_core()  )
		self.assertEqual( sccfg.get_env_config('dev').log_full_path , raw_cfg.dev.log_file()   )
	

	def test_0030_load_core_config_file(self):
		sccfg = CoreConfig( filename='sys_config.json', base_dir=os.getcwd() )

		raw_cfg = jsoncfg.load_config( 'sys_config.json' )
		# breakpoint()
		logger.debug( os.getcwd() )
		self.assertEqual( sccfg.get_env_config('dev').env , 'dev'  )
		self.assertEqual( sccfg.get_env_config('dev').host , raw_cfg.dev.host()  )
		self.assertEqual( sccfg.get_env_config('dev').port , raw_cfg.dev.port()  )
		self.assertEqual( sccfg.get_env_config('dev').log_file , raw_cfg.dev.log_file()  )
		self.assertEqual( sccfg.get_env_config('dev').db_core , raw_cfg.dev.db_core()  )
		self.assertEqual( sccfg.get_env_config('dev').log_full_path , raw_cfg.dev.log_file()   )
	

	def test_0040_load_core_config_path_str(self):

		sys_config2 = """{	
					"dev":{
						"host":"0.0.0.0",
						"port":"4100" ,
						"log_file":"!/log_test_dev.txt",
						"db_core":"postgresql:///test1"
					} 
			}"""

		sccfg = CoreConfig(config_str=sys_config2, base_dir=os.getcwd() )
		raw_cfg = jsoncfg.loads_config( sys_config2 )
		self.assertEqual( sccfg.get_env_config('dev').log_full_path , raw_cfg.dev.log_file().replace('!',  os.getcwd())   )

	def test_0050_flask_config(self):
		config = cdb.to_flask_config()
		self.assertTrue( 'SECRET_KEY' in config.keys() )
		self.assertTrue( 'SQLALCHEMY_TRACK_MODIFICATIONS' in config.keys() )

	def test_0060_query_get_onfig_xpath(self):
		self.assertTrue( cdb.get_config(config_str='SYS/DIR//ENV_BASE_DIR' ) != None )
		
	def test_0061_query_get_onfig_xpath(self):
		self.assertTrue( cdb.get_config(config_str='SYS///ENV_BASE_DIRxx' ) == None )

	def test_0062_query_get_onfig_xpath(self):
		self.assertTrue( cdb.get_config(config_str='SYS/x/x/ENV_BASE_DIR' ) == None )

	def test_0070_query_set_onfig_xpath(self):
		cdb.set_config(config_str='SYS/DIR//ENV_BASE_DIR', value='123' )
		self.assertTrue( cdb.get_config(config_str='SYS/DIR//ENV_BASE_DIR' ).value == '123' )

	def test_0071_query_set_onfig_xpath_new_item(self):
		cdb.set_config(config_str='SYS/DIR//LOG_FILE', value='log.txt' )
		self.assertTrue( cdb.get_config(config_str='SYS/DIR//LOG_FILE' ).value == 'log.txt' )
	
	def test_0072_query_set_onfig_xpath_new_config_and_item(self):
		cdb.set_config(config_str='MAIN///LOG_FILE', value='log.txt' )
		self.assertTrue( cdb.get_config(config_str='MAIN///LOG_FILE' ).value == 'log.txt' )


import unittest

import os , sys
# import sys, os, jsoncfg
sys.path.insert(0, '../../../')


from flask import Flask #, current_app
from mclogger import    MCLogger 
from pathlib import Path
from zohavi.zconfig.config_manager import CoreConfig

# base_dir = Path( os.getcwd() )
sccfg = CoreConfig( f'sys_config.json' )

from zohavi.zcore.appcore import AppCore

class TestCore(unittest.TestCase):

    def test_0010_start_app(self):
        app = Flask(__name__)
        myapp = AppCore(app, sccfg, 'dev' )

        port = myapp.get_config('SYS///PORT')
        # breakpoint()
        self.assertTrue( port.value == '4100')
		# self.assertTrue( myapp.get_config( config_str='SYS///LOG_FILE', value=sccfg.get_env_config( env ).log_file )
		# self._ref_config.set_config( config_str='SYS///HOST', value=sccfg.get_env_config( env ).host )
		# self._ref_config.set_config( config_str='SYS///PORT', value=sccfg.get_env_config( env ).port )
		# self._ref_config.set_config( config_str='SYS/DIR//BASE_DIR', value=sccfg.get_env_config( env ).base_dr )
		# self._ref_config.set_config( config_str='SYS/DB//CORE', value=sccfg.get_env_config( env ).db_core )


import unittest
unittest.TestLoader.sortTestMethodsUsing = None

import sys, os, jsoncfg
sys.path.insert(0, '../../../')


# from mclogger import MCLogger
from zohavi.zerrors.routes import ErrorsView 
from zohavi.zcore.appcore import AppCore 
# from zohavi.zconfig.models import MDL_Config
# from jsondbupload import JsonDBUpload

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_migrate import Migrate

from zohavi.zconfig.config_manager import CoreConfig

sys_config = """{
					"dev":{
						"host":"0.0.0.0",
						"port":"4100" ,
						"log_file":"log_test_dev.txt",
						"db_core":"postgresql:///test1"
					},
					"qa":{
						"host":"0.0.0.0",
						"port":"4200" ,
						"log_file":"log_test_qa.txt",
						"db_core":"postgresql:///test1"
					},
					"prd":{
						"host":"0.0.0.0",
						"port":"4300" ,
						"log_file":"log_test_prd.txt",
						"db_core":"postgresql:///test1"
					},
			}"""

sccfg = CoreConfig( config_str= sys_config  )


class TestCore(unittest.TestCase):

    def test_0010_start_app(self):
        app = Flask(__name__)
        myapp = AppCore(app, sccfg, 'dev' )

        myapp.setup_blueprint_dyn( 	bp_module_class_str ='zohavi.zerrors#bp', 
                                    route_module_class_str='zohavi.zerrors.routes#ErrorsView',
                                    url_prefix='/' )
        app.run(host="0.0.0.0", port=8601)


import unittest
unittest.TestLoader.sortTestMethodsUsing = None

import sys, os, jsoncfg
sys.path.insert(0, '../../../')
sys.path.insert(0, '../../../../')
sys.path.insert(0, '../../../../../')

from flask import Flask
from mclogger import MCLogger
from mclogger.mclogger import MCLogger

from zohavi.zcore.appcore import AppCore 
from zohavi.zdb.models import TBL_Default
from zohavi.zwebui.data_ui_model import DataUIModel


logger = MCLogger( 'test_log.txt' )

site_update_schema = {
							"SiteMain":[{ 	
								"module_name":"test_zwebui",
								"table_obj": "SiteMain",
								"fields":{
											"si_site_id":{"field_db":"id",   "key":True},
											"si_site_name":{"field_db":"site_name", "validation":{"required":True, "text_min_len":3, "text_max_len":20} },
											"si_site_code":{"field_db":"site_code", "validation":{"required":True,  "text_max_len":5} },
											"si_site_desc":{"field_db":"site_desc", "validation":{"required":False,  "text_max_len":200} }
								}
							}]
					}

from sqlalchemy import Column, Integer, String 
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

Base = declarative_base()

############################################################################################################
############################################################################################################
class SiteMain( Base ,  TBL_Default):
	__tablename__ = 'site_main' 
	id = Column(Integer(), primary_key=True)
	site_name = Column(String(100) ) 
	site_desc = Column(String(200) ) 
	site_code = Column(String(5) )  
	cfg_main_runme 	= Column(String(10) ) #dev, qa, prod  
# 	_env_list = relationship("SiteEnv", backref=backref("site_env" ), lazy='select') 
	
# ############################################################################################################
# ############################################################################################################
# class SiteEnv(Model,  TBL_Default): 
# 	ENV_STATUS_NEW = 'NEW'
# 	ENV_STATUS_GEN = 'GENERATING'
# 	ENV_STATUS_ACT = 'ACTIVE'
# 	ENV_STATUS_DIS = 'DISABLED'
# 	ENV_STATUS_DEL = 'DELETED'

# 	__tablename__ 	= 'site_env' 
# 	id 				= Column(Integer(), primary_key=True)
# 	site_id 		= Column(Integer(), ForeignKey( 'site_main.id'  ) )
# 	_site 			= relationship("SiteMain", backref=backref("site_main" ), lazy='joined')
# 	env_code 		= Column(String(5) ) #dev, qa, prod
# 	env_is_root		= Column(Boolean(), default=False )  
# 	env_desc		= Column(String(20) ) #dev, qa, prod
# 	env_setup_status = Column(String(10), default=ENV_STATUS_NEW ) #NEW, GENERATING, ACTIVE, DISABLED, DELETED

# 	prev_env_id		= Column(Integer(), ForeignKey( 'site_env.id'  ) )

class TestCore(unittest.TestCase):
	def setUp(self):

		engine = create_engine('sqlite:///:memory:')
		Session = sessionmaker(bind=engine)
		self.session = Session()

		Base.metadata.create_all(engine)
		self.session.add( SiteMain(site_name='twitterbot', site_desc='Twitter Bot', site_code='tb', cfg_main_runme='dev'))
		self.session.commit()

	########################################################################
	# Test creating a class
	def test_0001_validate_schema(self):
		logger.debug('.')
		validate_schema = {
							"SiteMain":[{ 	
								"module_name":"test_zwebui",
								"table_obj": "SiteMain",
								"fields":{
											"si_site_id":{"field_db":"id",   "key":True},
											"si_site_name":{"field_db":"site_name", "validation":{"required":True, "text_min_len":3, "text_max_len":5} },
											"si_site_code":{"field_db":"site_code", "validation":{"required":True,  "text_max_len":5} },
											"si_site_desc":{"field_db":"site_desc", "validation":{"required":False,  "text_max_len":200} }
								}
							}]
					}

		self.assertTrue(  DataUIModel.validate_schema( validate_schema["SiteMain"] )  )
		



	#######################################################################
	#Test creating a class
	def test_0010_create_instance(self):
		logger.debug('.')
		data_ui = DataUIModel( site_update_schema['SiteMain']  , self.session, logger ) 
		self.assertTrue( data_ui )
 		
	########################################################################
	# Test getting field info
	def test_0020_get_field_schema(self):
		logger.debug('.')
		data_ui = DataUIModel( site_update_schema['SiteMain']  , self.session, logger ) 

		site_name = data_ui.get_field_schema('SiteMain', "si_site_name")
		print( site_name )
		self.assertTrue( site_name )

	########################################################################
	# Test updating fieleds - create records with validated fields
	def test_0025_create_record(self):
		logger.debug('.')
		data_ui = DataUIModel( site_update_schema['SiteMain']  , self.session, logger ) 
		json_update = 	[
							{ "id":"si_site_id", "value":""},
							{ "id":"si_site_name", "value":"abc1"},
							{ "id":"si_site_desc", "value":"abc2"},
							{ "id":"si_site_code", "value":"abc3"}
						]
		ret_data = data_ui.data_update_ajax( json_update )

		item = self.session.query( SiteMain).filter_by( site_name='abc1' ).first()
		self.assertEquals( ret_data[1], 200  )
		
		
		self.assertTrue( item  )

	########################################################################
	# Test updating fieleds - create records with failed validated fields
	def test_0025_create_record(self):
		logger.debug('.')
		data_ui = DataUIModel( site_update_schema['SiteMain']  , self.session, logger ) 
		json_update = 	[
							{ "id":"si_site_id", "value":""},
							{ "id":"si_site_name", "value":"ccccccccccccccccccccccccccccccccccccc"},
							{ "id":"si_site_desc", "value":"abc2"},
							{ "id":"si_site_code", "value":"abc3"}
						]
		ret_data = data_ui.data_update_ajax( json_update )
		# breakpoint()
		logger.debug( ret_data )

		item = self.session.query( SiteMain).filter_by( site_name='abc1' ).first()
		self.assertEquals( ret_data[1], 500  )
		self.assertFalse( item  )

	########################################################################
	# Test updating fieleds
	def test_0030_update_record(self):
		logger.debug('.')
		data_ui = DataUIModel( site_update_schema['SiteMain']  , self.session, logger ) 
		json_update = 	[
							{ "id":"si_site_id", "value":""},
							{ "id":"si_site_name", "value":"abc1"},
							{ "id":"si_site_desc", "value":"abc2"},
							{ "id":"si_site_code", "value":"abc3"}
						]
		ret_data = data_ui.data_update_ajax( json_update )
		logger.debug( ret_data )
		item = self.session.query( SiteMain).filter_by( site_name='abc1' ).first()
		self.assertEquals( ret_data[1], 200  )
		self.assertTrue( item  )

	########################################################################
	# Test updating fieleds but iwht missing fields
	def test_0040_update_record_with_missing_fields(self):
		logger.debug('.')
		data_ui = DataUIModel( site_update_schema['SiteMain']  , self.session, logger ) 
		json_update = 	[
							#{ "id":"si_site_id", "value":""}, <--- omit to test failures
							# { "id":"si_site_name", "value":"abc1"},
							# { "id":"si_site_desc", "value":"abc2"},
							# { "id":"si_site_code", "value":"abc3"}
						]
		ret_data = data_ui.data_update_ajax( json_update )
		item = self.session.query( SiteMain).filter_by( site_name='abc1' ).first()
		 
		self.assertEquals( ret_data[1], 500  )
		self.assertFalse( item  )

	########################################################################
	# Test updating fieleds but iwht missing fields
	

#  #################################################################
#  #################################################################


if __name__ == '__main__':
    unittest.main()


import smtplib, ssl
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.header    import Header

from prepost import PreCond 
from zohavi.zbase.staple import ZStaple

class Emailer(ZStaple):
	####################################################

	@PreCond.dict_has_fields( config=[ 'HOST', 'PORT', 'SENDER', 'PASSWORD'])
	def __init__(self, config, logger=None):
		super().__init__(logger = logger )
		self.debug = True

		self.config  = config
		self.context = ssl.create_default_context()

	# @celeryJob.task(name="cjob.send_email")
	def send_email(self, subject, sender, recipients, text_body, html_body=None):
		with smtplib.SMTP_SSL( self.config['HOST'], self.config['PORT'], self.context ) as server:
			server.login( self.config['SENDER'] , self.config['PASSWORD'])

			if(html_body):
				msg = MIMEMultipart( "alternative" )
				msg.attach(  MIMEText(text_body, "plain") )
				msg.attach(  MIMEText(html_body, "html") )
			else:
				msg = MIMEText( text_body, 'plain', 'utf-8')
			
			msg['Subject'] = Header('email test', 'utf-8')
			msg['From'] = sender
			msg['To'] = ", ".join(  recipients )
			

			return server.sendmail( sender , recipients , msg.as_string() )

	# ####################################################
	# def generate_confirmation_token(email):
	# 	serializer = URLSafeTimedSerializer(self.config['SECRET_KEY'])
	# 	return serializer.dumps(email, salt=self.config['SECURITY_PASSWORD_SALT'])

	# ####################################################
	# def confirm_token(token, expiration=3600):
	# 	serializer = URLSafeTimedSerializer(self.config['SECRET_KEY'])
	# 	try:
	# 		email = serializer.loads(
	# 			token,
	# 			salt=self.config['SECURITY_PASSWORD_SALT'],
	# 			max_age=expiration
	# 		)
	# 	except:
	# 		return False
	# 	return email

	# ####################################################
	# # @celeryJob.task(name="cjob.send_email")
	# def send_register_email_confirmation(user):
	# 	token = generate_confirmation_token(user.email)
	# 	send_email.apply_async( args={ 'subject':' Confrim Your email',
	# 				'sender':self.config['MAIL_USERNAME'],
	# 				'recipients': [user.email],
	# 				'text_body':  render_template('zemail/confirm_email.txt',
	# 										user=user, token=token),
	# 				'html_body': render_template('zemail/confirm_email.html',
	# 										user=user, token=token) } )   


	



import  json 
from jsonschema import validate as json_schema_validate
from jsonschema.exceptions import ValidationError


from zohavi.zwebui.web_field import WebField
from zohavi.zbase.staple import ZStaple
# from _def.common.selector import Selector
from dataobjtools.selector import Selector
import importlib

class DataUIModel(ZStaple):
	#####################################################################################
	def __init__(self,   form_field_validation_schema, session, logger ):

		super().__init__(app = None, logger = logger )

		DataUIModel.validate_schema( form_field_validation_schema )

		# self.update_obj_list = update_obj
		self.data_schema = form_field_validation_schema 
		
		self.web_schema = WebField( form_field_validation_schema, logger=logger )
		# self.logger = logger
		self.session = session

		# pass
	@staticmethod
	def validate_schema( json_input_schema):
		validation_schema = {
							"type": "array",
							"items": 
							{ 	"type" : "object",
								"properties" : 
								{ "module_name":{"type":"string" },
									"table_obj":{"type":"string" },
									"fields": 
									{ "type": "object",
										"patternProperties": 
										{	".*": 
											{ 	"type": "object",
												"properties": 
												{ 	"field_db":{"type":"string" },
													"key":{"type":"boolean" },
													"validation":{
																	"type": "object",
																	"properties": {
																					"required":{"type":"boolean"},
																					"text_min_len":{"type":"integer"},
																					"text_max_len":{"type":"integer"}
								} 	}	}	}	}	}	}	}
							}
		try:
			json_schema_validate( instance= json_input_schema , schema=validation_schema   )
		except  Exception as e:
			print(e)
			return False

		return True
 
	#####################################################################################
	def get_webfield_value( self, web_data, field_html_name):
		for field in web_data:	
			if field[ 'id'] == field_html_name: return field['value']
		
		return None

	#####################################################################################
	def get_field_schema( self, obj_name, field_html_name):
		
		for table_obj_rec in self.data_schema:
			if obj_name == table_obj_rec['table_obj']:
				return table_obj_rec['fields'].get(field_html_name)


		# table_obj_rec = self.data_schema.get( obj_name )

		# if table_obj_rec:
			
		return None

	#####################################################################################
	def _data_validate( self, submit_data ): 
		validation_result = {}
		self.log_debug("validating data")
		self.log_debug( submit_data )

		validation_result = self.web_schema.validate(  submit_data )
		if validation_result['success']: self.log_debug( 'Validation checks completed - result: passed' )
		else: self.log_error('Validation checks completed - result: failed')
		self.log_debug("****:" + json.dumps(   validation_result ) )
		
		return validation_result

	#####################################################################################
	def data_delete_ajax( self, submit_data ): 
		validation_result = self._data_validate( submit_data)
		if validation_result['success']:
			data_obj_list = self.data_delete( submit_data ) 
			return json.dumps({'success':True, 'data':  data_obj_list , 'result': validation_result, 'schema':self.data_schema }), 200
		else:
			return json.dumps({'success':False , 'result': validation_result }), 500 

	#####################################################################################
	def data_get_ajax( self, search_dict ):  
		
		data_obj_list = self.data_get( search_dict )
		
		if data_obj_list: return json.dumps({'success':True, 'data':  data_obj_list , 'schema':self.data_schema }), 200 

		return json.dumps({'success':False}), 500  

	#####################################################################################
	def data_get( self, search_dict, ret_db_obj=False):
		data_obj_list = []
		## Example of talbe fields in self.data_schema
		# 	"SiteMain":{ 	
		# 		# "table":"SiteMain",
		# 		# "fields":{
		# 					"si_site_id":{"field_db":"site_id",   "key":True},
		# 					"si_site_name":{"field_db":"site_name", "validation":{"required":True, "text_min_len":3, "text_max_len":20} },
		# 					"si_site_code":{"field_db":"site_code",   "validation":{"required":True,  "text_max_len":5} },
		# 		# } }
		#Loop through each of the json table schame descriptions
		for obj_name, data_table in self.data_schema.items():
			# if obj_name in self.update_obj_list:
			db_fields = self.data_get_table_fields( data_table['fields'] , search_dict )
			self.log_debug(f"{obj_name} Read table: {data_table['module_name']}::{ data_table['table_obj'] } with search keys {db_fields['keys'] }" )	
			
			if db_fields:
				data_class_ref = getattr(importlib.import_module( data_table['module_name'] ), data_table['table_obj'] )	#Get ref to table object name dynamically
				data_obj = self.session.query( data_class_ref ).filter_by( **db_fields['keys'] ).all()

				for data_item in data_obj:
					if ret_db_obj: data_obj_list.append( data_item  )
					else: data_obj_list.append( data_item.to_dict()  )

		[ self.log_debug( f'Data from query: {data_item}')  for data_item in data_obj_list ] 
		return data_obj_list

	#####################################################################################
	def data_update_ajax( self, submit_data ): 
		if not submit_data:
			self.log_error('No data given')  
			return json.dumps({'success':False}), 500 
		else:
			validation_result = self._data_validate( submit_data)
			if validation_result['success']:
				data_obj_list = self.data_update( submit_data )
				# breakpoint()
				# webfield_mapping = self.data_get_web_fields( self.data_schema, data_obj_list)
				return json.dumps({'success':True, 'data':  data_obj_list , 'result': validation_result, 'schema':self.data_schema  }), 200
				# return json.dumps({'success':True, 'data':  data_obj_list , 'schema':self.data_schema  }), 200
			else:
				self.log_error('Could not update data as valdation failed')  
				return json.dumps({'success':False,  'result': validation_result } ), 500 



	#####################################################################################
	def data_delete(self, web_data):
		for data_table in self.data_schema:
			db_fields = self.data_get_table_fields( data_table['fields'] , web_data )
			self.log_debug(f" Delete table: {data_table['module_name']}::{ data_table['table_obj'] } with search keys {db_fields['keys'] }" )

			data_class_ref = getattr(importlib.import_module( data_table['module_name'] ), data_table['table_obj']  )	#Get ref to table object name dynamically
			data_obj = None
			if db_fields['keys']: 
				data_obj = self.session.query( data_class_ref ).filter_by( **db_fields['keys'] ).delete()
				self.session.commit()
		return True
			# breakpoint()
	#####################################################################################
	def data_update(self, web_data):
		data_obj_list = []
		# breakpoint()
		## Example of talbe fields in self.data_schema
		# 	"SiteMain":{ 	
		# 		# "table":"SiteMain",
		# 		# "fields":{
		# 					"si_site_id":{"field_db":"site_id",   "key":True},
		# 					"si_site_name":{"field_db":"site_name", "validation":{"required":True, "text_min_len":3, "text_max_len":20} },
		# 					"si_site_code":{"field_db":"site_code",   "validation":{"required":True,  "text_max_len":5} },
		# 		# } }
		#Loop through each of the json table schame descriptions
		for data_table in self.data_schema :
			# if obj_name in self.update_obj_list:
			db_fields = self.data_get_table_fields( data_table['fields'] , web_data )
			self.log_debug(f"Modify table: {data_table['module_name']}::{ data_table['table_obj'] } with search keys {db_fields['keys'] }" )	
			
			data_class_ref = getattr(importlib.import_module( data_table['module_name'] ), data_table['table_obj'] )	#Get ref to table object name dynamically
			data_obj = None

			#Search to see if the record exists
			if db_fields['keys']: 
				# breakpoint()
				data_obj = self.session.query( data_class_ref ).filter_by( **db_fields['keys'] ).first()
			
			if data_obj:  #If found, then update fields
				for db_field_name in db_fields['fields']:
					setattr( data_obj, db_field_name, db_fields['fields'][ db_field_name ]  )
			else:	#If not found, create new object
				data_obj = data_class_ref( **db_fields['fields'] )
				self.session.add(data_obj)  
				self.session.flush()	#specifically flush to get the ID number
			data_obj_list.append( data_obj.to_dict()  )
			# else: 
			# 	self.log_debug(f"Skipping [{obj_name}] as it is not in the schema list provided[{self.update_obj_list}]")
		
		self.session.commit()
		
		return data_obj_list 


	#####################################################################################
	def data_get_table_fields(self, table_schema_fields, web_data ):
		data_obj_list = []
		db_fields = {'keys':{}, 'fields':{} } 
		# breakpoint()
		for field_name, field in table_schema_fields.items():	#Loop through each of the fields in a given table
			
			new_field = Selector().dict_search( web_data, {'id': field_name  } )
			if not new_field: 
				self.log_error(f'Could not find key [{field_name}] within web_data:[{web_data}]')
			else:
				# self.logger.debug(f"Adding field [{field_name}] value of: [{ new_field['value']  }]")
				if field.get("key") :
					if new_field['value']: db_fields[ "keys"][ field["field_db"] ] = new_field['value'] 
				else:
					db_fields[ "fields"][ field["field_db"] ] = new_field['value'] 

		self.log_debug(f"fields: {db_fields}")
		return db_fields 

 

# class DataUIModel(ZStaple):
# 	#####################################################################################
# 	def __init__(self, update_obj, form_field_validation_schema, session, logger ):

# 		super().__init__(app = None, logger = logger )

# 		self.update_obj_list = update_obj
# 		self.data_schema = form_field_validation_schema 
# 		self.web_schema = WebField( form_field_validation_schema, logger=logger )
# 		# self.logger = logger
# 		self.session = session

# 		# pass

# 	#####################################################################################
# 	def get_webfield_value( self, web_data, field_html_name):
# 		for field in web_data:	
# 			if field[ 'id'] == field_html_name: return field['value']
		
# 		return None

# 	#####################################################################################
# 	def get_field_schema( self, obj_name, field_html_name):
# 		table_obj_rec = self.data_schema.get( obj_name )

# 		if table_obj_rec:
# 			return table_obj_rec['fields'].get(field_html_name)
# 		return None

# 	#####################################################################################
# 	def _data_validate( self, submit_data ): 
# 		self.log_debug("validating data")
# 		self.log_debug( submit_data )

# 		if self.web_schema.validate(  submit_data ):
# 			self.log_debug('validation ok')
# 			return True
# 		return False

# 	#####################################################################################
# 	def data_delete_ajax( self, submit_data ): 
# 		if self._data_validate( submit_data):
# 			data_obj_list = self.data_delete( submit_data ) 
# 			return json.dumps({'success':True, 'data':  data_obj_list , 'schema':self.data_schema }), 200
# 		else:
# 			self.log_error('valdation failed')  
# 			return json.dumps({'success':False}), 500 

# 	#####################################################################################
# 	def data_get_ajax( self, search_dict ):  
		
# 		data_obj_list = self.data_get( search_dict )
		
# 		if data_obj_list: return json.dumps({'success':True, 'data':  data_obj_list , 'schema':self.data_schema }), 200 

# 		return json.dumps({'success':False}), 500  

# 	#####################################################################################
# 	def data_get( self, search_dict, ret_db_obj=False):
# 		data_obj_list = []
# 		## Example of talbe fields in self.data_schema
# 		# 	"SiteMain":{ 	
# 		# 		# "table":"SiteMain",
# 		# 		# "fields":{
# 		# 					"si_site_id":{"field_db":"site_id",   "key":True},
# 		# 					"si_site_name":{"field_db":"site_name", "validation":{"required":True, "text_min_len":3, "text_max_len":20} },
# 		# 					"si_site_code":{"field_db":"site_code",   "validation":{"required":True,  "text_max_len":5} },
# 		# 		# } }
# 		#Loop through each of the json table schame descriptions
# 		for obj_name, data_table in self.data_schema.items():
# 			if obj_name in self.update_obj_list:
# 				db_fields = self.data_get_table_fields( data_table['fields'] , search_dict )
# 				self.log_debug(f"{obj_name} Read table: {data_table['module_name']}::{ data_table['table_obj'] } with search keys {db_fields['keys'] }" )	
				
# 				if db_fields:
# 					data_class_ref = getattr(importlib.import_module( data_table['module_name'] ), data_table['table_obj'] )	#Get ref to table object name dynamically
# 					data_obj = self.session.query( data_class_ref ).filter_by( **db_fields['keys'] ).all()

# 					for data_item in data_obj:
# 						if ret_db_obj: data_obj_list.append( data_item  )
# 						else: data_obj_list.append( data_item.to_dict()  )

# 		[ self.log_debug( f'Data from query: {data_item}')  for data_item in data_obj_list ] 
# 		return data_obj_list

# 	#####################################################################################
# 	def data_update_ajax( self, submit_data ): 
# 		if not submit_data:
# 			self.log_error('No data given')  
# 			return json.dumps({'success':False}), 500 
# 		elif self._data_validate( submit_data):
# 			data_obj_list = self.data_update( submit_data )
# 			# breakpoint()
# 			# webfield_mapping = self.data_get_web_fields( self.data_schema, data_obj_list)
# 			return json.dumps({'success':True, 'data':  data_obj_list , 'schema':self.data_schema  }), 200
# 			# return json.dumps({'success':True, 'data':  data_obj_list , 'schema':self.data_schema  }), 200
# 		else:
# 			self.log_error('valdation failed')  
# 			return json.dumps({'success':False}), 500 



# 	#####################################################################################
# 	def data_delete(self, web_data):
# 		for obj_name, data_table in self.data_schema.items():
# 			db_fields = self.data_get_table_fields( data_table['fields'] , web_data )
# 			self.log_debug(f"{obj_name} Delete table: {data_table['module_name']}::{ data_table['table_obj'] } with search keys {db_fields['keys'] }" )

# 			data_class_ref = getattr(importlib.import_module( data_table['module_name'] ), data_table['table_obj']  )	#Get ref to table object name dynamically
# 			data_obj = None
# 			if db_fields['keys']: 
# 				data_obj = self.session.query( data_class_ref ).filter_by( **db_fields['keys'] ).delete()
# 				self.session.commit()
# 		return True
# 			# breakpoint()
# 	#####################################################################################
# 	def data_update(self, web_data):
# 		data_obj_list = []
# 		# breakpoint()
# 		## Example of talbe fields in self.data_schema
# 		# 	"SiteMain":{ 	
# 		# 		# "table":"SiteMain",
# 		# 		# "fields":{
# 		# 					"si_site_id":{"field_db":"site_id",   "key":True},
# 		# 					"si_site_name":{"field_db":"site_name", "validation":{"required":True, "text_min_len":3, "text_max_len":20} },
# 		# 					"si_site_code":{"field_db":"site_code",   "validation":{"required":True,  "text_max_len":5} },
# 		# 		# } }
# 		#Loop through each of the json table schame descriptions
# 		for obj_name, data_table in self.data_schema.items():
# 			if obj_name in self.update_obj_list:
# 				db_fields = self.data_get_table_fields( data_table['fields'] , web_data )
# 				self.log_debug(f"{obj_name} Modify table: {data_table['module_name']}::{ data_table['table_obj'] } with search keys {db_fields['keys'] }" )	
				
# 				data_class_ref = getattr(importlib.import_module( data_table['module_name'] ), data_table['table_obj'] )	#Get ref to table object name dynamically
# 				data_obj = None

# 				#Search to see if the record exists
# 				if db_fields['keys']: data_obj = self.session.query( data_class_ref ).filter_by( **db_fields['keys'] ).first()
				
# 				if data_obj:  #If found, then update fields
# 					for db_field_name in db_fields['fields']:
# 						setattr( data_obj, db_field_name, db_fields['fields'][ db_field_name ]  )
# 				else:	#If not found, create new object
# 					data_obj = data_class_ref( **db_fields['fields'] )
# 					self.session.add(data_obj)  
# 					self.session.flush()	#specifically flush to get the ID number
# 				data_obj_list.append( data_obj.to_dict()  )
# 			else: 
# 				self.log_debug(f"Skipping [{obj_name}] as it is not in the schema list provided[{self.update_obj_list}]")
		
# 		self.session.commit()
		
# 		return data_obj_list 


# 	#####################################################################################
# 	def data_get_table_fields(self, table_schema_fields, web_data ):
# 		data_obj_list = []
# 		db_fields = {'keys':{}, 'fields':{} } 
# 		# breakpoint()
# 		for field_name, field in table_schema_fields.items():	#Loop through each of the fields in a given table
			
# 			new_field = Selector().dict_search( web_data, {'id': field_name  } )
# 			if not new_field: 
# 				self.log_error(f'Could not find key [{field_name}] within web_data:[{web_data}]')
# 			else:
# 				# self.logger.debug(f"Adding field [{field_name}] value of: [{ new_field['value']  }]")
# 				if field.get("key") :
# 					if new_field['value']: db_fields[ "keys"][ field["field_db"] ] = new_field['value'] 
# 				else:
# 					db_fields[ "fields"][ field["field_db"] ] = new_field['value'] 

# 		self.log_debug(f"fields: {db_fields}")
# 		return db_fields 


# 	# #####################################################################################
# 	# def data_get_web_fields(self, table_schema_fields, db_data ):
# 	# 	data_obj_list = []
# 	# 	web_fields = { } 

# 	# 	for schema in table_schema_fields:
# 	# 		for field_name in table_schema_fields[schema]['fields']:
# 	# 			field_rec = table_schema_fields[schema]['fields'][field_name]
# 	# 			if field_rec['field_db'] in db_data[0]:		#See if the database field name is in the data row.  
# 	# 														#No need to check all rows, just check the first row 
# 	# 														#since all rows schema is the same
# 	# 				web_fields[ field_rec['field_db'] ]  =  field_name
# 	# 	return web_fields 


import json




def add_item( validation: dict  ):

    validation['success'] = validation['success'] and False 

    if 'validation' not in validation: validation['validation'] = []
    validation['validation'].append( { 'rule':'is_url', 'value':'23', 'result':False } )


validation = { 'success':True}
print("BEFORE")
print( json.dumps( validation ))
print("AFTER")
add_item(validation)
print( json.dumps( validation ))



import os, sys, ast
from dataclasses import dataclass, field
from typing import List, Dict 


import jsoncfg

from prepost import PreCond
from jsondbupload import JsonDBUpload

@dataclass
class CoreConfigItem():

	@property
	def log_full_path(self) -> str:
		return self.log_file.replace( "!", self.base_dir)

	env:str
	host: str
	port: int
	db_core: str 
	base_dir: str
	log_file: str
	log_full_path: str= field(init=False, default=log_full_path)


class CoreConfig():
	@PreCond.minimum_not_null(2)
	def __init__(self, filename:str=None, config_str:str=None, base_dir:str=None):
		self.env_config = {}

		if filename: config = config = jsoncfg.load_config( filename )
		elif config_str: config = config = jsoncfg.loads_config( config_str )
		self._extract_values( config , base_dir)

	def _extract_values(self, config_raw_data_dict, base_dir:str):
		# breakpoint()
		for config_env in config_raw_data_dict().keys():
			data_dict = {}
			# for config_item in config_env['config_list']:
			data_dict['env'] = config_env

			env_data = config_raw_data_dict[ config_env ]()
			for field  in env_data.keys():
				data_dict[ field ] = env_data[ field ]

			data_dict['base_dir'] = base_dir
			scc_item = CoreConfigItem( **data_dict ) 
			self.env_config[ config_env ] = scc_item
			# print(scc_item)

	def get_env_config(self, env) -> CoreConfigItem:
		return self.env_config[ env ] 




############################################################
############################################################
############################################################
############################################################
class ConfigLoaderFunc():
	@staticmethod
	def append_to_env_field(cfg, raw_param):
		param = ast.literal_eval(raw_param) 
		return ConfigLoaderFunc().get_env_field(cfg, raw_param) + param['suffix'] 

	@staticmethod
	def get_env_field(cfg, raw_param):
		param = ast.literal_eval(raw_param) 
		curr_config = cfg.get_env_config( param['key_path']) #[0], param['key'][1], param['key'][2], param['key'][3]  )
		return curr_config 

	@staticmethod
	def get_working_dir(cfg,  raw_param):
		curr_dir = os.path.dirname( os.path.abspath(sys.argv[0]) )
		return curr_dir + '/'		

	@staticmethod
	def get_random_key(cfg,   raw_param):
		return os.urandom(32)

	@staticmethod
	def load_config( cfg, raw_param):
		path = ConfigLoaderFunc().append_to_env_field(cfg, raw_param)
		print( f"***checking {path}")
		config = jsoncfg.load_config(path)()

		param = ast.literal_eval(raw_param) 
		# breakpoint()
		if 'load_sub_path' in param:
			for token in param['load_sub_path'].split('/'):
				config = config[token]
		# breakpoint()
		return config

# ############################################################
# class ConfigManager():
# 	def __init__(self, session, logger):
# 		self.session = session
# 		self.logger = logger

# 		self.J2DB = JsonDBUpload( session, logger)
# 		# pass

# 	#####################################################################################
# 	#	Save the config data to database
# 	def save_defaults(self, config_defaults_file):
# 		self.J2DB.update_tables_from_file( config_defaults_file )
# 		# pass

# 	#####################################################################################
# 	#	Read the config data
# 	def load(self):
# 		pass

# 	def get(self, area=None, category=None, sub_cat=None, name=None, default=None):
# 		pass

# ############################################################
# class ConfigCoreLoader():
# 	@PreCond.minimum_not_null(1)
# 	def __init__(self, filename=None, data_str = None):
# 		if filename: self._config = jsoncfg.load_config( filename)
# 		elif data_str: self._config = jsoncfg.loads_config( data_str )
	
# 	def get_config(self):
# 		return self._config
# 	def get_config_env(self, env, config_item):
# 		return self._config[env][config_item].value

############################################################
class ConfigTemplate():
	CFG = {} 
	env_code = "def"

	############################################################
	def __init__(self, env_code ):
		self.env_code = env_code

	# ############################################################
	# def set_lib_path(self): 
	# 	filedir = os.path.dirname( os.path.abspath(__file__) )		#get current directory
	# 	basedir = "/".join( filedir.split("/")[:-1] )				#get path above current directory
	# 	sys.path.append( basedir  )									#this would be the app dir
	# 	parentdir = "/".join( filedir.split("/")[:-3] )				#get path above current directory
		
	# 	sys.path.append(  parentdir )

	############################################################
	def get_env_config(self, xpath):  #area/category/sub_cat/name, or area/category//name, or area///name
		token = xpath.split('/')  
		if not 'ENV' in self.CFG: self.CFG['ENV'] = {}
		config = self.CFG['ENV'][self.env_code]
		for token_item in token:
			if not token_item in config: raise Exception( f"Path {xpath} [{token_item}] not found in current ENV[{self.env_code}]")  
			config = config[ token_item ]
		return config

	############################################################
	def get_app_config(self, xpath):
		token = xpath.split('/')  
		config = self.CFG['APP'] 
		for token_item in token:
			config = config[ token_item ]
		return config
		# breakpoint()

	############################################################
	def get_cfg(self, name, area, cat="_"    ):
			return self.CFG[ area ][cat][name]

	############################################################
	def get_env(self, name, area, cat="_"    ):
			return self.ENV.CFG[ area ][cat][name]

	############################################################
	def get_app(self, name, area, cat="_"    ):
			return self.APP.CFG[ area ][cat][name]

	##############################################################################################################
	# Load configuration 
	def load_app_config(self,  config_file ): 	
		raw_config = jsoncfg.load_config( config_file) 

		self.CFG[ 'APP' ] = {} 
		for raw_config_entry in raw_config[ 'app_config_list' ]:
			 self._load_config_entry(raw_config_entry, self.CFG[ 'APP' ])
		
		return self

	##############################################################################################################
	# Load configuration 
	def load_env_config(self, config_file ): 
		raw_config = jsoncfg.load_config( config_file) 
		
		for raw_env_config in raw_config[ 'env_config_list' ]:
			self.CFG[ 'ENV' ] = {}
			self.CFG[ 'ENV' ][ raw_env_config.env_code() ]  = {}
			cfg_path = self.CFG[ 'ENV' ][ raw_env_config.env_code() ]

			for raw_cfg_grp in raw_env_config['CFG']:	#Go through each file config entry
				self._load_config_entry(raw_cfg_grp, cfg_path)
		return self

	##############################################################################################################
	# Load configuration 
	def _load_config_entry(self, raw_cfg_grp , cfg_path):
		#Defaul t the config AREA, CATEGORY, SUB-CATEGORY if provided and create empty dict if equired
		if not raw_cfg_grp.area() in cfg_path: cfg_path[  raw_cfg_grp.area() ] = {}
		if not raw_cfg_grp.category() in cfg_path[  raw_cfg_grp.area() ]:  cfg_path[  raw_cfg_grp.area() ][ raw_cfg_grp.category() ] = {}
		if not raw_cfg_grp.sub_cat() in cfg_path[  raw_cfg_grp.area() ][ raw_cfg_grp.category() ]: cfg_path[  raw_cfg_grp.area() ][ raw_cfg_grp.category() ][ raw_cfg_grp.sub_cat() ] = {}

		#Get reference to the root item
		cfg_item = cfg_path[  raw_cfg_grp.area() ][ raw_cfg_grp.category() ][ raw_cfg_grp.sub_cat() ]
		
		#For each config item, then process - ether call func, convert jason, get value
		for raw_cfg_item in raw_cfg_grp[ 'config_list']:
			if 'func' in raw_cfg_item:	
				func_name =  raw_cfg_item.func()
				result = getattr(ConfigLoaderFunc,  func_name )(self, raw_cfg_item.param( {} ) )
				cfg_item[ raw_cfg_item.name() ] = result
			elif 'value_json' in raw_cfg_item: 
				cfg_item[ raw_cfg_item.name() ] = ast.literal_eval( raw_cfg_item.value_json(None) )
			else:
				cfg_item[ raw_cfg_item.name() ] = raw_cfg_item.value(None) 


	################################################################################################
	# Convert key config fields to dictionary to support flask input
	def _convert_to_flask(self, flask_list, obj_list):		#get key fields to flask
		flask_config = {}
		for item in flask_list:
			config_tokens = item.split('/')
			# breakpoint()
			for obj in obj_list:
				flask_config = {**flask_config, **self._to_flask_get_token( obj, 0, config_tokens) }
		
		flask_config['obj'] = self

		return flask_config

	##########################################################################################
	# Flatten the dictionary
	def _to_flask_get_token(self, obj, index, config_tokens):
		ret_hash = {}

		curr_token = config_tokens[index]	#Get the current token to process

		if index < ( len(config_tokens)-1): 	#If we still have more tokens
			if curr_token == "*":			#If current token is an '*', then loop through and merge with other items
				for key in obj:		
					ret_hash = {**ret_hash , **self._to_flask_get_token( obj[key], index+1, config_tokens) }
				return ret_hash
			else:
				next_node = obj.get( curr_token, None)
				if next_node: return {**ret_hash , **self._to_flask_get_token( next_node, index+1, config_tokens)}
				return ret_hash
		else:	#If this is last token, then copy over current records
			if curr_token == "*":
				for key in obj:
					ret_hash[ key] = obj[key] 
				return ret_hash
			else:
				return obj[ curr_token ]
				 


#IMPORT STANDARD
import os, sys , re
 
from flask import Flask, url_for 
from flask_sqlalchemy import SQLAlchemy
from flask_migrate import Migrate
from flask_login import LoginManager 

import pymongo
from zohavi.zwebui.jinja_cust_funcs import JCFunc
from zohavi.zconfig.models import MDL_Config
	
class AppCore:
	def __init__(self, app,  sccfg, env, test_only=False): 
		self._app = app 
		# self._app.config.update( config.to_flask()  )

		self._app.config['SQLALCHEMY_DATABASE_URI'] = sccfg.get_env_config( env ).db_core
		self.db = SQLAlchemy(self._app)
		self._ref_config = MDL_Config( logger=None, db=self.db)

		if not test_only: 
			self._sync_core_config_to_db( sccfg, env )
			self._app.config.update( self._ref_config.to_flask_config()  )

		# self.db = SQLAlchemy(self._app)
		self._migrate = Migrate(self._app, self.db)


		self.login = LoginManager(self._app)
		self.login.login_message_category = "danger"  #when cannot login, set error type to red color on display
		self.login.login_view = 'user.login'		#The function name in routes that will handle the login page 

		# self._mongodb = pymongo.MongoClient(   self.get_config( "MONGODB_URL" ) ) #"mongodb://localhost:27017/"

		self._set_custom_template_loaders()
		self._define_template_funcions( self._app.jinja_env )

	def run(self, *args, **kwargs):
		self._app.run(*args, **kwargs)

	##############################################################################################
	# Copy the basic cofig and put it to the database
	def _sync_core_config_to_db(self,  sccfg, env):
		self._ref_config.set_config( config_str='SYS///LOG_FILE', value=sccfg.get_env_config( env ).log_file )
		self._ref_config.set_config( config_str='SYS///HOST', value=sccfg.get_env_config( env ).host )
		self._ref_config.set_config( config_str='SYS///PORT', value=sccfg.get_env_config( env ).port )
		self._ref_config.set_config( config_str='SYS/DIR//BASE_DIR', value=sccfg.get_env_config( env ).base_dir )
		self._ref_config.set_config( config_str='SYS/DB//CORE', value=sccfg.get_env_config( env ).db_core )


	##############################################################################################
	def get_config(self, config_str):
		# breakpoint()
		return self._ref_config.get_config( config_str=config_str)

	################################################################################################
	# Add reference to the main logger
	def register_logger(self, logger):
		self._logger = logger

	# ################################################################################################
	# # Commit to the database with an exception 
	# def safe_add_and_commit(self, obj ):
	# 	self.db.session.add(obj)
	# 	return self.safe_commit(obj )

	# ################################################################################################
	# # Commit to the database with an exception 
	# def safe_commit(self, obj=None ):
	# 	try:
	# 		self.db.session.commit()
	# 		if obj: 
	# 			self.db.session.refresh(obj) 
	# 			return obj
	# 	except Exception as error:
	# 		self._logger.error("DB Error:" + str(error)  )
	# 	return True
		 
	##############################################################################################################
	def _set_custom_template_loaders(self):
		from jinja2 import ChoiceLoader, FileSystemLoader

		if 'CUST_TEMPLATE_OVERRIDEDIR' in self._app.config:
			my_loader = ChoiceLoader([ FileSystemLoader(self._app.config['CUST_TEMPLATE_OVERRIDEDIR']) ])
			self._app.jinja_loader = my_loader

	##############################################################################################################
	def _define_template_funcions(self, jinja_env):
		for method in dir(JCFunc):
			if method.startswith('__') is False:
				jinja_env.globals[ 'jc_' + method] = getattr(JCFunc, method)
		 
	##############################################################################################################
	def setup_blueprint_dyn( self, bp_module_class_str ,  route_module_class_str =None,  url_prefix =None  ):

		#import the module that contains the FlaskView class
		bp_tokens = bp_module_class_str.split('#')
		if len(bp_tokens) != 2: raise Exception('Must pass module#class_name - e.g. app.siteadmin.routes#SiteAdminView')
		bp_module = __import__( bp_tokens[0] , globals(), locals(), [ bp_tokens[1] ] ) 
		bp_class =  getattr(bp_module, bp_tokens[1] )

		#call the register function from the FlaskView class to ensure the routes are registered
		if route_module_class_str: 
			route_tokens = route_module_class_str.split('#')
			if len(route_tokens) != 2: raise Exception('Must pass module#class_name - e.g. app.siteadmin.routes#SiteAdminView')
			route_module = __import__( route_tokens[0] , globals(), locals(), [ route_tokens[1] ] )  
			route_class = getattr( route_module, route_tokens[1] )
			# breakpoint()
			route_class.register(  bp_class  )	

		# if local_templates: blueprint.bp.jinja_loader.searchpath.append( local_templates )
		

		self._app.register_blueprint( bp_class , url_prefix= url_prefix   )

		# return blueprint.bp



import json
from pathlib import Path
  
from flask_classful import  FlaskView,  route 
from . import bp

from flask_login import current_user #,   login_required
from flask_classful import FlaskView, route
from flask import   send_file, current_app, render_template,  current_app , abort, jsonify

from .staple import ZStaple
 


class BaseView(FlaskView, ZStaple):
	route_base = '/'
	# bpname = 'base'
	bp = bp

	##############################################################################################################
	def __init__(self):
		self.module_name = __name__ 

	##############################################################################################################
	def register_app(self, app_ref, logger_ref):
		self.myapp = app_ref
		self.logger = logger_ref

	# def log_debug(self, message):
	# 	if self.logger: self.logger.debug( message )
	# 	else: print( "DEBUG:" + message )

	# def log_error(self, message):
	# 	if self.logger: self.logger.error( message )
	# 	else: print( "ERROR:" + message )

	# def log_info(self, message):
	# 	if self.logger: self.logger.info( message )
	# 	else: print( "INFO:" + message )

	# def log_warning(self, message):
	# 	if self.logger: self.logger.warning( message )
	# 	else: print( "WARNING:" + message )


	##############################################################################################################
	#	list all routes
	##############################################################################################################
	@route("/routes", methods=["GET"])
	def getRoutes(self):
		routes = {}
		for rule_item in current_app.url_map._rules:
			routes[rule_item.rule] = {}
			routes[rule_item.rule]["functionName"] = rule_item.endpoint
			routes[rule_item.rule]["methods"] = list(rule_item.methods)

		# routes.pop("/static/<path:filename>")

		return jsonify(routes)


	# @self.myapp.app.after_request
	# def set_response_headers(response):
	# 	response.headers['Cache-Control'] = 'no-cache, no-store, must-revalidate'
	# 	response.headers['Pragma'] = 'no-cache'
	# 	response.headers['Expires'] = '0'
	# 	return response

	##############################################################################################################
	#	Return any static resources
	##############################################################################################################
	@route('/st/<string:subdir_type>/<string:module>/<path:resourceFile>')
	def get_page(self, subdir_type, module, resourceFile):		
		# path = current_app.config['APP_BASE_DIR']
		path = current_app.config['ENV_BASE_DIR']
		if subdir_type in [ "_def", "app"]: 
			path += subdir_type + "/" 
		else:
			logger.error( f"incorrect type sent of:{subdir_type}")
			abort(500)

		send_file_path = path + module + "/" + resourceFile

 
		if Path(send_file_path).is_file():
			logger.debug("file found:" + str(send_file_path)   )
			return send_file(send_file_path)

		logger.error("file missing:" + str(send_file_path)   )
		abort(404)




"""empty message

Revision ID: 3c00080e3336
Revises: 
Create Date: 2022-04-18 01:54:36.181293

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3c00080e3336'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zohavi',
 'zohavi.zbase',
 'zohavi.zcelery',
 'zohavi.zcommon',
 'zohavi.zconfig',
 'zohavi.zconfig.tests',
 'zohavi.zconfig.tests.migrations',
 'zohavi.zconfig.tests.migrations.versions',
 'zohavi.zcore',
 'zohavi.zcore.tests',
 'zohavi.zdb',
 'zohavi.zemailer',
 'zohavi.zerrors',
 'zohavi.zerrors.tests',
 'zohavi.zmembers',
 'zohavi.zwebui',
 'zohavi.zwebui.tests']

package_data = \
{'': ['*'],
 'zohavi.zbase': ['templates/zbase/*'],
 'zohavi.zcelery': ['templates/_def/*'],
 'zohavi.zemailer': ['templates/zemail/*'],
 'zohavi.zerrors': ['static/_def/images/*', 'templates/zerrors/*'],
 'zohavi.zmembers': ['templates/zmembers/*'],
 'zohavi.zwebui': ['static/zcss/*', 'static/zjs/*', 'static/zwc/*'],
 'zohavi.zwebui.tests': ['templates/*']}

install_requires = \
['json-cfg', 'pymongo']

setup_kwargs = {
    'name': 'zohavi',
    'version': '0.1.47',
    'description': 'Web widgets',
    'long_description': '# Hello World 123456789',
    'author': 'pub12',
    'author_email': 'pubudu79@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import unittest
unittest.TestLoader.sortTestMethodsUsing = None

from pathlib import Path 
import sys, os, jsoncfg, json
sys.path.insert(0, '../../../')


from zohavi.zwebui.jinja_cust_funcs import JCFunc


class TestCore(unittest.TestCase):

	########################################################################
	# Test
	def test_0010_defaulter(self):
		# jc = JCFunc()

		value = JCFunc.defaulter( 'abc', 'def')
		self.assertEqual(  value , 'abc')
 		
		value = JCFunc.defaulter( None, 'def')
		self.assertEqual(  value , 'def')

		value = JCFunc.defaulter( '', 'def')
		self.assertEqual(  value , 'def')

	########################################################################
	# Test
	def test_0020_dict_to_json_str(self):
		dict_out = {'name':'john', 'address':'NY', 'age':22, 'gender':'M'}

		json_str = JCFunc.dict_to_json_str( dict_out , ['apple'] )
		self.assertEqual(  json_str , '{}' )

		json_str = JCFunc.dict_to_json_str( dict_out , ['name'] )
		self.assertEqual(  json_str , '{"name": "john"}' )
 		

		json_str = JCFunc.dict_to_json_str( dict_out , ['name', 'age'] )
		self.assertEqual(  json_str , '{"name": "john", "age": 22}' )
 		
	########################################################################
	# Test
	def test_0030_list_dict_to_json_str(self):
		dict_out = [ {'name':'john', 'address':'NY', 'age':22, 'gender':'M'},
					 {'name':'jill', 'address':'HK', 'age':44, 'gender':'F'},
					 {'name':'jack', 'address':'CN', 'age':14, 'gender':'M'} ]

		json_str = JCFunc.list_dict_to_json_str( dict_out , ['name'] )
		self.assertEqual(  json_str , '[{"name": "john"}, {"name": "jill"}, {"name": "jack"}]' )
 		
		json_str = JCFunc.list_dict_to_json_str( dict_out , ['name', 'age'] )
		self.assertEqual(  json_str , '[{"name": "john", "age": 22}, {"name": "jill", "age": 44}, {"name": "jack", "age": 14}]' )
 		
		json_str = JCFunc.list_dict_to_json_str( dict_out , ['apple'] )
		self.assertEqual(  json_str , '[]' )
 		
 		

if __name__ == '__main__':
    unittest.main()



from flask import  render_template, request  
# from flask_classful import   route

# from app import myapp, logger #,  db 
from . import bp  
from zohavi.zbase import BaseView
 

class ErrorsView(BaseView):
	# route_base = '/'
	bpname = bp.name
	bp = bp
	
@bp.app_errorhandler(404)
def not_found_error(error):
	# logger.debug('404 page not found for:{}'.format(request.url))
	return render_template('zerrors/404.html'), 404


@bp.app_errorhandler(500)
def internal_error(error):
	# myapp.db.session.rollback()
	return render_template('zerrors/500.html'), 500
